(******************************************************************************)
(* computer (stack machine)                                                   *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

IMPORT FIO;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  acMax = 32;
  vcMax = 32;

VAR
  args: ARRAY [ 1 .. acMax ] OF TValue;  (* stack *)
  ac: CARDINAL;
  vars: ARRAY [ 0 .. vcMax - 1 ] OF TValue;  (* memory *)
  r0, r1: TValue;

PROCEDURE Dbg0();
BEGIN
  FIO.WrStr( dbg, 'Comp.Debug: ac       top-1         top' );
  FIO.WrLn( dbg );
END Dbg0;

PROCEDURE DbgH( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( dbg, 'Comp.Debug: ' );
  FIO.WrCard( dbg, ac, 2 );
  FIO.WrChar( dbg, ' ' );
  IF ac = 0 THEN
    FIO.WrStr( dbg, '                       ' );
  ELSIF ac = 1 THEN
    FIO.WrStr( dbg, '            ' );
    FIO.WrInt( dbg, args[ac], 11 );
  ELSE
    FIO.WrInt( dbg, args[ac-1], 11 );
    FIO.WrChar( dbg, ' ' );
    FIO.WrInt( dbg, args[ac], 11 );
  END;
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgH;

PROCEDURE Dbg( s: ARRAY OF CHAR );
BEGIN
  DbgH( s );
  FIO.WrLn( dbg );
END Dbg;

PROCEDURE DbgV( s: ARRAY OF CHAR; v: TValue );
BEGIN
  DbgH( s );
  FIO.WrStr( dbg, ' ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgO( s: ARRAY OF CHAR; o: TOffset );
BEGIN
  DbgH( s );
  FIO.WrStr( dbg, ' ' );
  FIO.WrCard( dbg, o, 1 );
  FIO.WrLn( dbg );
END DbgO;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ErrOvr();
BEGIN
  Err( 'Arithmetic overflow' );
END ErrOvr;

PROCEDURE Idx( o: TOffset ): CARDINAL;
BEGIN
  IF o >= vcMax THEN Err( 'Invalid offset' ) END;
  RETURN o;
END Idx;

PROCEDURE Push();
BEGIN
  IF ac >= acMax THEN Err( 'Stack overflow' ) END;
  INC( ac ); args[ ac ] := r0;
END Push;

PROCEDURE PopX( VAR x: TValue );
BEGIN
  IF ac <= 0 THEN Err( 'Stack underflow' ) END;
  x := args[ ac ]; DEC( ac );
END PopX;

PROCEDURE Pop();
BEGIN
  PopX( r0 );
END Pop;

PROCEDURE Pop2();
BEGIN
  PopX( r1 ); PopX( r0 );
END Pop2;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN DbgV( 'Val', v ) END;
  r0 := v; Push();
END Val;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN DbgO( 'Lod', o ) END;
  r0 := vars[ Idx(o) ]; Push();
END Lod;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN DbgO( 'Sto', o ) END;
  Pop(); vars[ Idx(o) ] := r0;
END Sto;

PROCEDURE Opr( o: TOperation );
BEGIN
  CASE o OF
  | oWrt: IF debug THEN Dbg( 'Wrt' ) END;
          Pop(); FIO.WrInt( out, r0, 0 ); FIO.WrLn( out );
  | oNeg: IF debug THEN Dbg( 'Neg' ) END;
          Pop(); r0 := -r0; Push();
  | oAdd: IF debug THEN Dbg( 'Add' ) END;
          Pop2(); r0 := r0 + r1; Push();
  | oSub: IF debug THEN Dbg( 'Sub' ) END;
          Pop2(); r0 := r0 - r1; Push();
  | oMul: IF debug THEN Dbg( 'Mul' ) END;
          Pop2(); r0 := r0 * r1; Push();
  | oDiv: IF debug THEN Dbg( 'Div' ) END;
          Pop2(); IF r1 = 0 THEN Err( 'Division by zero' ) END;
          r0 := r0 DIV r1; Push();
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

BEGIN
  ac := 0;
  IF debug THEN Dbg0() END;
END Comp.
